
Imports System.Collections
Imports System.Collections.Generic
Imports System.Diagnostics

Namespace Generic

    ' ----------------------------------------------------------- a simple double-ended-queue collection of objects
    <Serializable()> _
    Partial Public Class Deque(Of T)
        Implements ICollection
        Implements IEnumerable(Of T)
        Implements ICloneable


        ' ------------------------------------------------------- node at the front of the deque
        Private front As Node = Nothing
        ' ------------------------------------------------------- node at the back of the deque
        Private back As Node = Nothing
        ' ------------------------------------------------------- number of elements in the deque
        Private m_count As Integer = 0
        ' ------------------------------------------------------- version of the deque
        Private version As Long = 0

        Public Sub New()
        End Sub

        Public Sub New(ByVal collection As IEnumerable(Of T))
            If collection Is Nothing Then
                Throw New ArgumentNullException("col")
            End If

            For Each item As T In collection
                PushBack(item)
            Next
        End Sub



        ' ---------------------------------------------------------------- removes all objects from the Deque
        '
        Public Overridable Sub Clear()
            m_count = 0

            front = back

            version += 1

            AssertValid()
        End Sub


        ' -------------------------------------------------------- test if element is in the Deque.
        '
        Public Overridable Function Contains(ByVal obj As T) As Boolean
            For Each o As T In Me
                If EqualityComparer(Of T).[Default].Equals(o, obj) Then
                    Return True
                End If
            Next
            Return False
        End Function


        ' -------------------------------------------------------- inserts an object at the front of the Deque
        '
        Public Overridable Sub PushFront(ByVal item As T)
            ' The new node to add to the front of the deque.
            Dim newNode As New Node(item)

            ' Link the new node to the front node. The current front node at 
            ' the front of the deque is now the second node in the deque.
            newNode.[Next] = front

            ' If the deque isn't empty.
            If Count > 0 Then
                ' Link the current front to the new node.
                front.Previous = newNode
            End If

            ' Make the new node the front of the deque.
            front = newNode

            ' Keep track of the number of elements in the deque.
            m_count += 1

            ' If this is the first element in the deque.
            If Count = 1 Then
                ' The front and back nodes are the same.
                back = front
            End If

            version += 1

            AssertValid()
        End Sub


        ' --------------------------------------------------------- inserts an object at the back of the Deque
        '
        Public Overridable Sub PushBack(ByVal item As T)
            ' The new node to add to the back of the deque.
            Dim newNode As New Node(item)

            ' Link the new node to the back node. The current back node at 
            ' the back of the deque is now the second to the last node in the
            ' deque.
            newNode.Previous = back

            ' If the deque is not empty.
            If Count > 0 Then
                ' Link the current back node to the new node.
                back.[Next] = newNode
            End If

            ' Make the new node the back of the deque.
            back = newNode

            ' Keep track of the number of elements in the deque.
            m_count += 1

            ' If this is the first element in the deque.
            If Count = 1 Then
                ' The front and back nodes are the same.
                front = back
            End If

            version += 1

            AssertValid()
        End Sub


        ' ------------------------------------------------- removes and returns the object at the front of the Deque
        '
        Public Overridable Function PopFront() As T
            If Count = 0 Then
                Throw New InvalidOperationException("Deque is empty.")
            End If

            ' Get the object at the front of the deque.
            Dim item As T = front.Value

            ' Move the front back one node.
            front = front.[Next]

            ' Keep track of the number of nodes in the deque.
            m_count -= 1

            ' If the deque is not empty.
            If Count > 0 Then
                ' Tie off the previous link in the front node.
                front.Previous = Nothing
            Else
                ' Else the deque is empty.
                ' Indicate that there is no back node.
                back = Nothing
            End If

            version += 1

            AssertValid()

            Return item
        End Function


        ' -------------------------------------------------- removes and returns the object at the back of the Deque
        '
        Public Overridable Function PopBack() As T
            If Count = 0 Then
                Throw New InvalidOperationException("Deque is empty.")
            End If

            ' Get the object at the back of the deque.
            Dim item As T = back.Value

            ' Move back node forward one node.
            back = back.Previous

            ' Keep track of the number of nodes in the deque.
            m_count -= 1

            ' If the deque is not empty.
            If Count > 0 Then
                ' Tie off the next link in the back node.
                back.[Next] = Nothing
            Else
                ' Else the deque is empty.
                ' Indicate that there is no front node.
                front = Nothing
            End If

            version += 1

            AssertValid()

            Return item
        End Function


        ' ------------------------------------------ returns the object at the front of the Deque without removing it
        '
        Public Overridable Function PeekFront() As T
            If Count = 0 Then
                Throw New InvalidOperationException("Deque is empty.")
            End If
            Return front.Value
        End Function


        ' ------------------------------------------ returns the object at the back of the Deque without removing it
        '
        Public Overridable Function PeekBack() As T
            If Count = 0 Then
                Throw New InvalidOperationException("Deque is empty.")
            End If
            Return back.Value
        End Function

        ' --------------------------------------------- Copies the Deque to a new array.
        Public Overridable Function ToArray() As T()
            Dim array As T() = New T(Count - 1) {}
            Dim index As Integer = 0

            For Each item As T In Me
                array(index) = item
                index += 1
            Next

            Return array
        End Function

        ' ---------------------------------------------- Returns a synchronized (thread-safe) wrapper for the Deque.
        '
        Public Shared Function Synchronized(ByVal deque As Deque(Of T)) As Deque(Of T)
            If deque Is Nothing Then
                Throw New ArgumentNullException("deque")
            End If
            Return New SynchronizedDeque(deque)
        End Function

        <Conditional("DEBUG")> _
        Private Sub AssertValid()
            Dim n As Integer = 0
            Dim current As Node = front

            While current IsNot Nothing
                n += 1
                current = current.[Next]
            End While

            Debug.Assert(n = Count)

            If Count > 0 Then
                Debug.Assert(front IsNot Nothing AndAlso back IsNot Nothing, "Front/Back Null Test - Count > 0")

                Dim f As Node = front
                Dim b As Node = back

                While f.[Next] IsNot Nothing AndAlso b.Previous IsNot Nothing
                    f = f.[Next]
                    b = b.Previous
                End While

                Debug.Assert(f.[Next] Is Nothing AndAlso b.Previous Is Nothing, "Front/Back Termination Test")
                Debug.Assert(f Is back AndAlso b Is front, "Front/Back Equality Test")
            Else
                Debug.Assert(front Is Nothing AndAlso back Is Nothing, "Front/Back Null Test - Count == 0")
            End If
        End Sub



#Region "ICollection Members"


        Public Overridable ReadOnly Property IsSynchronized() As Boolean Implements ICollection.IsSynchronized
            Get
                Return False
            End Get
        End Property

        Public Overridable ReadOnly Property Count() As Integer Implements ICollection.Count
            Get
                Return m_count
            End Get
        End Property


        ' --------------------------------------- Copies the Deque elements to an existing one-dimensional Array, 
        '                                         starting at the specified array index.
        '
        Public Overridable Sub CopyTo(ByVal array As Array, ByVal index As Integer) Implements ICollection.CopyTo
            If array Is Nothing Then
                Throw New ArgumentNullException("array")
            ElseIf index < 0 Then
                Throw New ArgumentOutOfRangeException("index", index, "Index is less than zero.")
            ElseIf array.Rank > 1 Then
                Throw New ArgumentException("Array is multidimensional.")
            ElseIf index >= array.Length Then
                Throw New ArgumentException("Index is equal to or greater " & "than the length of array.")
            ElseIf Count > array.Length - index Then
                Throw New ArgumentException("The number of elements in the source Deque is greater " & "than the available space from index to the end of the " & "destination array.")
            End If

            Dim i As Integer = index

            For Each obj As Object In Me
                array.SetValue(obj, i)
                i += 1
            Next
        End Sub


        ' ------------------------------------ gets an object that can be used to synchronize access to the Deque
        '
        Public Overridable ReadOnly Property SyncRoot() As Object Implements ICollection.SyncRoot
            Get
                Return Me
            End Get
        End Property

#End Region

#Region "IEnumerable Members"

        ' -------------------------------------- Returns an enumerator that can iterate through the Deque.
        '
        Private Function IEnumerable_GetEnumerator() As IEnumerator Implements IEnumerable.GetEnumerator
            Return New Enumerator(Me)
        End Function

#End Region

#Region "ICloneable Members"

        ' ---------------------------------------------------------------- creates a shallow copy of the Deque
        '
        Public Overridable Function Clone() As Object Implements ICloneable.Clone
            Dim _clone As New Deque(Of T)(Me)

            _clone.version = Me.version

            Return _clone
        End Function

#End Region

#Region "IEnumerable<T> Members"

        Public Overridable Function GetEnumerator() As IEnumerator(Of T) Implements IEnumerable(Of T).GetEnumerator
            Return New Enumerator(Me)
        End Function

#End Region

    End Class

End Namespace
