

Imports System.Diagnostics
Imports System.IO

Public Enum SmpteFrameRate
	Smpte24 = 24
	Smpte25 = 25
	Smpte30Drop = 29
	Smpte30 = 30
End Enum

Public Enum SequenceType
	Ppqn
	Smpte
End Enum


Friend Class MidiFileProperties
	Private Const PropertyLength As Integer = 2

    ' TODO - LIVIO - Asc is OK ???
    Private Shared ReadOnly MidiFileHeader As Byte() = {Asc("M"c), Asc("T"c), Asc("h"c), Asc("d"c), 0, 0, 0, 6}
	Private m_format As Integer = 1
	Private m_trackCount As Integer = 0
	Private m_division As Integer = PpqnClock.PpqnMinValue
	Private m_sequenceType As SequenceType = SequenceType.Ppqn

	Public Sub New()
	End Sub

	Public Sub Read(strm As Stream)
        If strm Is Nothing Then
            Throw New ArgumentNullException("strm")
        End If

        m_format = InlineAssignHelper(m_trackCount, InlineAssignHelper(m_division, 0))

        FindHeader(strm)
        Format = ReadProperty(strm)
        TrackCount = ReadProperty(strm)
        Division = ReadProperty(strm)

        AssertValid()
    End Sub

	Private Sub FindHeader(stream As Stream)
		Dim found As Boolean = False
		Dim result As Integer

		While Not found
			result = stream.ReadByte()

            If result = Asc("M"c) Then
                result = stream.ReadByte()

                If result = Asc("T") Then
                    result = stream.ReadByte()

                    If result = Asc("h"c) Then
                        result = stream.ReadByte()

                        If result = Asc("d"c) Then
                            found = True
                        End If
                    End If
                End If
            End If

			If result < 0 Then
				Throw New MidiFileException("Unable to find MIDI file header.")
			End If
		End While

        ' ------------------------------------------------------------------ eat the header length
		For i As Integer = 0 To 3
			If stream.ReadByte() < 0 Then
				Throw New MidiFileException("Unable to find MIDI file header.")
			End If
		Next
	End Sub

	Private Function ReadProperty(strm As Stream) As UShort
		Dim data As Byte() = New Byte(PropertyLength - 1) {}

		Dim result As Integer = strm.Read(data, 0, data.Length)

		If result <> data.Length Then
			Throw New MidiFileException("End of MIDI file unexpectedly reached.")
		End If

		If BitConverter.IsLittleEndian Then
			Array.Reverse(data)
		End If

		Return BitConverter.ToUInt16(data, 0)
	End Function

	Public Sub Write(strm As Stream)
        If strm Is Nothing Then
            Throw New ArgumentNullException("strm")
        End If

        strm.Write(MidiFileHeader, 0, MidiFileHeader.Length)
        WriteProperty(strm, CUShort(Format))
        WriteProperty(strm, CUShort(TrackCount))
        WriteProperty(strm, CUShort(Division))
    End Sub

	Private Sub WriteProperty(strm As Stream, [property] As UShort)
		Dim data As Byte() = BitConverter.GetBytes([property])

		If BitConverter.IsLittleEndian Then
			Array.Reverse(data)
		End If

		strm.Write(data, 0, PropertyLength)
	End Sub

	Private Shared Function IsSmpte(division As Integer) As Boolean
		Dim result As Boolean
		Dim data As Byte() = BitConverter.GetBytes(CShort(division))

		If BitConverter.IsLittleEndian Then
			Array.Reverse(data)
		End If

		If CSByte(data(0)) < 0 Then
			result = True
		Else
			result = False
		End If

		Return result
	End Function

	<Conditional("DEBUG")> _
	Private Sub AssertValid()
		If m_trackCount > 1 Then
			Debug.Assert(Format = 1 OrElse Format = 2)
		End If

		If IsSmpte(Division) Then
			Debug.Assert(SequenceType = SequenceType.Smpte)
		Else
			Debug.Assert(SequenceType = SequenceType.Ppqn)
			Debug.Assert(Division Mod PpqnClock.PpqnMinValue = 0)
		End If
	End Sub

	Public Property Format() As Integer
		Get
			Return m_format
		End Get
		Set
            If Value < 0 OrElse Value > 3 Then
                Throw New ArgumentOutOfRangeException("Format", Value, "MIDI file format out of range.")
            ElseIf Value = 0 AndAlso m_trackCount > 1 Then
                Throw New ArgumentException("MIDI file format invalid for this track count.")
            End If

            m_format = Value

            AssertValid()
        End Set
	End Property

	Public Property TrackCount() As Integer
		Get
			Return m_trackCount
		End Get
		Set
            If Value < 0 Then
                Throw New ArgumentOutOfRangeException("TrackCount", Value, "Track count out of range.")
            ElseIf Value > 1 AndAlso Format = 0 Then
                Throw New ArgumentException("Track count invalid for this format.")
            End If

            m_trackCount = Value

            AssertValid()
        End Set
	End Property

	Public Property Division() As Integer
		Get
			Return m_division
		End Get
		Set
			If IsSmpte(value) Then
				Dim data As Byte() = BitConverter.GetBytes(CShort(value))

				If BitConverter.IsLittleEndian Then
					Array.Reverse(data)
				End If

                If CSByte(data(0)) <> -CInt(SmpteFrameRate.Smpte24) AndAlso CSByte(data(0)) <> -CInt(SmpteFrameRate.Smpte25) AndAlso CSByte(data(0)) <> -CInt(SmpteFrameRate.Smpte30) AndAlso CSByte(data(0)) <> -CInt(SmpteFrameRate.Smpte30Drop) Then
                    Throw New ArgumentException("Invalid SMPTE frame rate.")
                Else
                    m_sequenceType = SequenceType.Smpte
                End If
			Else
				If value Mod PpqnClock.PpqnMinValue <> 0 Then
					Throw New ArgumentException("Invalid pulses per quarter note value.")
				Else
					m_sequenceType = SequenceType.Ppqn
				End If
			End If

			m_division = value

			AssertValid()
		End Set
	End Property

	Public ReadOnly Property SequenceType() As SequenceType
		Get
			Return m_sequenceType
		End Get
    End Property

	Private Shared Function InlineAssignHelper(Of T)(ByRef target As T, value As T) As T
		target = value
		Return value
    End Function

End Class


Public Class MidiFileException
    Inherits ApplicationException

	Public Sub New(message As String)
		MyBase.New(message)
    End Sub

End Class
